/* Copyright © Nils O.Selåsdal <NOS@Utel.no> 
   Comments/bug reports/anyting goes to NOS@Utel.no
   License is BSD, see the LICENSE file
*/
#include <gtk/gtk.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <openssl/evp.h>
#include <openssl/opensslv.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include "ghasher.h"
#include "hash.xpm"


static gboolean actionbutton_sum(GtkButton *button,struct GHApp *app);
static gboolean actionbutton_cancel(GtkButton *button,struct GHApp *app);

const struct ActionButton actions[] = {
	{"_Digest",actionbutton_sum},
	{"_Cancel",actionbutton_cancel},
	{NULL,NULL}
};

static gint action = 0;

//event/source id if running calculation(idle function)
static gint current_idle_id = -1;
const char *const hashfunctions[] = {
	 "md2",
	 "md4",
	 "md5",
	 "sha1",
	 "sha",
	 "ripemd160",
	 "dss1"
};
//called when the calculation(idle function) is finished
static void md5_idle_destroy(gpointer p)
{
	struct SumContext *s_ctx = p;
	int md_len;
	char b[EVP_MAX_MD_SIZE];
	char b_hex[EVP_MAX_MD_SIZE*2+1];
	close(s_ctx->currentfile);
	current_idle_id = -1;
	md_len = 0;
//openssl prior to 0.9.7 needs a Final else we leak memory afaik,
//above 0.9.7 openssl provides a _cleanup function.
#if OPENSSL_VERSION_NUMBER < 0x00907000L
	EVP_DigestFinal(&s_ctx->ctx,b,&md_len);
#endif
	
	if(s_ctx->allok){
#if OPENSSL_VERSION_NUMBER >= 0x00907000L
			EVP_DigestFinal(&s_ctx->ctx,b,&md_len);
#endif
			hex_dump(b,md_len,b_hex,md_len*2+1);
			gtk_entry_set_text(GTK_ENTRY(s_ctx->app->sumentry),b_hex);
			gtk_progress_bar_set_text(GTK_PROGRESS_BAR(s_ctx->app->progress),"Digesting finished");

	}
#if OPENSSL_VERSION_NUMBER >= 0x00907000L
	EVP_MD_CTX_cleanup(&s_ctx->ctx);
#endif
	gtk_widget_set_sensitive(GTK_WIDGET(s_ctx->app->fileentry),TRUE);
	gtk_widget_set_sensitive(GTK_WIDGET(s_ctx->app->filebutton),TRUE);
	gtk_widget_set_sensitive(GTK_WIDGET(s_ctx->app->digestdropdown),TRUE);
	g_free(s_ctx);
}

//idle function. Runs when no other events are available and calculates a piece of the file,
//then gives control back to the gui.
static gboolean md5_idle(gpointer p)
{
	struct SumContext *s_ctx = p;
	unsigned char buffer[s_ctx->statbuf.st_blksize*2];
	gboolean finished = FALSE;
	int cnt = 50;
	while(cnt--){
		ssize_t ret;

		ret = read(s_ctx->currentfile,buffer,s_ctx->statbuf.st_blksize*2);
		if(ret > 0){
			EVP_DigestUpdate(&s_ctx->ctx,buffer,(unsigned long)ret);	
			s_ctx->read_acc += (off_t)ret;
		}else if(ret == 0){
			s_ctx->allok = TRUE;
			finished = TRUE;
        		gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(s_ctx->app->progress),1.0);
		} else if(ret == -1){
			if(errno == EINTR){
				return TRUE;
			}
			show_error_dialog(s_ctx->app,"Digesting failed\n%s",strerror(errno));
			finished = TRUE;
		}
		if(finished){
			action = 1 - action;
			gtk_button_set_label(GTK_BUTTON(s_ctx->app->actionbutton),actions[action].name);
			return FALSE;

		}
			
	}

        gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(s_ctx->app->progress),(float)s_ctx->read_acc/(float)s_ctx->statbuf.st_size);
	return TRUE;
}

static void actionbutton_clicked(GtkButton *button,gpointer userdata)
{
	struct GHApp *app = userdata;
	if(actions[action].callback(button,app)){
		action = 1 - action;
		gtk_button_set_label(button,actions[action].name);
	}
}

static gboolean actionbutton_sum(GtkButton *button,struct GHApp *app)
{
	struct SumContext *s_ctx;
	const gchar *filename;
	
	s_ctx = g_malloc0(sizeof(*s_ctx));
	filename = gtk_entry_get_text(GTK_ENTRY(app->fileentry));
	if(stat(filename,&s_ctx->statbuf)== -1){
		g_free(s_ctx);
		show_error_dialog(app,"Could not open '%s' :\n%s",filename,strerror(errno));// everyone ofcourse understands strerror() messages... 
		return FALSE;
	}
	
	s_ctx->md = EVP_get_digestbyname(app->digesttype);
	if(s_ctx->md == NULL) {
		g_free(s_ctx);
		show_error_dialog(app,"Digest type '%s' is not supported.",app->digesttype);
		return FALSE;

	}

	EVP_DigestInit(&s_ctx->ctx,s_ctx->md);
	s_ctx->currentfile = open(filename,O_RDONLY);
	if(s_ctx->currentfile < 0){
		g_free(s_ctx);
		show_error_dialog(app,"Could not open '%s' :\n%s",filename,strerror(errno));
		return FALSE;
	}

	s_ctx->allok = FALSE;
	s_ctx->app = app;
	current_idle_id = g_idle_add_full(G_PRIORITY_DEFAULT_IDLE,md5_idle,s_ctx,md5_idle_destroy);	
	gtk_progress_bar_set_text(GTK_PROGRESS_BAR(s_ctx->app->progress),"Digesting...");
	gtk_entry_set_text(GTK_ENTRY(s_ctx->app->sumentry),"");
	gtk_widget_set_sensitive(GTK_WIDGET(app->fileentry),FALSE);
	gtk_widget_set_sensitive(GTK_WIDGET(app->filebutton),FALSE);
	gtk_widget_set_sensitive(GTK_WIDGET(s_ctx->app->digestdropdown),FALSE);
	return TRUE;
}

static gboolean actionbutton_cancel(GtkButton *button, struct GHApp *app)
{
	if(current_idle_id != -1){//wtf ?
		g_source_remove(current_idle_id);
		gtk_progress_bar_set_text(GTK_PROGRESS_BAR(app->progress),"Digesting cancelled");
	}

	return TRUE;
}

static void filenamebutton_clicked(GtkButton *button,gpointer userdata)
{
	struct GHApp *app = userdata;
	GtkWidget *file_selector;
	gint ret;
	const gchar *filename;

	filename = gtk_entry_get_text(GTK_ENTRY(app->fileentry));

   	/* Create the selector , use the new gtkfilechooser if gtk >= 2.4.0*/
   
#if GTK_CHECK_VERSION(2,4,0)
	file_selector = gtk_file_chooser_dialog_new ("Select a file to digest. ",GTK_WINDOW(app->window),GTK_FILE_CHOOSER_ACTION_OPEN,GTK_STOCK_OPEN, GTK_RESPONSE_OK,GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,NULL);
	if(g_path_is_absolute(filename)){ //seems to go mickey if not an absolute path
		gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(file_selector),filename); 
	}
	 gtk_file_chooser_set_local_only(GTK_FILE_CHOOSER(file_selector),TRUE); //we only handle kernel filesystems
#else

   	file_selector = gtk_file_selection_new ("Select a file to digest.");
  	gtk_window_set_transient_for(GTK_WINDOW(file_selector),GTK_WINDOW(app->window)); 
	gtk_file_selection_set_filename(GTK_FILE_SELECTION(file_selector),filename);
#endif
   
   	ret = gtk_dialog_run (GTK_DIALOG(file_selector));
	if(ret ==  GTK_RESPONSE_OK){
#if GTK_CHECK_VERSION(2,4,0)
		gtk_entry_set_text(GTK_ENTRY(app->fileentry),gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(file_selector)));
#else
		gtk_entry_set_text(GTK_ENTRY(app->fileentry),gtk_file_selection_get_filename(GTK_FILE_SELECTION(file_selector)));
#endif
		gtk_entry_set_text(GTK_ENTRY(app->sumentry),""); //reset previous digest
	}

	gtk_widget_destroy(file_selector);
}

static void setup_window_icon(GtkWidget *window)
{
	GdkPixbuf *icon;
	icon =  gdk_pixbuf_new_from_xpm_data((const char**)hash_xpm);
	if(icon != NULL){
		gtk_window_set_icon(GTK_WINDOW(window),icon);
		g_object_unref(icon);
	}

}

static void print_usage()
{
	g_print("Copyright © Nils O.Selåsdal <NOS@Utel.no> \n");
	g_print("Usage: ghasher [-h] [-d type] [filename]\n");
	g_print("\t-d calculate 'type' digest where 'type' is md2|md4|md5|sha1|sha|ripemd160|dss1\n");
	g_print("\t-h this text\n");
	g_print("\tfilename name of file to digest\n");
}

static void parse_options(int argc, char *argv[],struct GHApp *app)
{
	int c;
	
	app->digesttype = "md5"; //default is md5
	app->cmdfilename = NULL;

	while ((c = getopt(argc, argv, "d:h")) != -1) {
		switch(c){
			case 'h':
				print_usage();
				exit(1);
			case 'd':
				app->digesttype = optarg;
				break;
			case '?':
				break;
			default:
				g_print("Unknown option %c\n",c);
				break;
		}
	}
	if(optind < argc){
		app->cmdfilename = argv[optind];
	}
}

static gboolean start_hash(gpointer userdata)
{
	struct GHApp *app = userdata;

	g_assert(app->cmdfilename != NULL);
	gtk_entry_set_text(GTK_ENTRY(app->fileentry),app->cmdfilename);
	g_signal_emit_by_name(G_OBJECT(app->actionbutton),"clicked",app,NULL);

	return TRUE;
}
#if GTK_CHECK_VERSION(2,4,0)
static void digest_changed(GtkComboBox *box,gpointer userdata)
{
	struct GHApp *app = userdata;
	gint index;
	g_object_get(G_OBJECT(box),"active",&index,NULL);
	app->digesttype = hashfunctions[index];

	gtk_entry_set_text(GTK_ENTRY(app->sumentry),"");

}

static GtkWidget *hashdropdown(struct GHApp *app)
{
	GtkWidget *combo;
	gint i;

	combo = gtk_combo_box_new_text();
	for(i = 0 ; i < sizeof(hashfunctions)/sizeof(hashfunctions[0]) ; i++){
		gtk_combo_box_append_text(GTK_COMBO_BOX(combo),hashfunctions[i]);
		if(!g_ascii_strncasecmp(app->digesttype,hashfunctions[i],MIN(strlen(app->digesttype),strlen(hashfunctions[i])))){
			g_object_set(G_OBJECT(combo),"active",i,NULL);
		}
	}

	g_signal_connect(G_OBJECT(combo),"changed",G_CALLBACK(digest_changed),app);

	return combo;
}
#else 
struct MenuHelper{
	struct GHApp *app;
	int hashindex;
};

static void digest_changed(GtkMenuItem *box,gpointer userdata)
{
	struct MenuHelper *mh = userdata;
	struct GHApp *app = mh->app;

	app->digesttype = hashfunctions[mh->hashindex];
	gtk_entry_set_text(GTK_ENTRY(app->sumentry),"");
}

static GtkWidget *hashdropdown(struct GHApp *app)
{
	GtkWidget *combo,*menu;
	gint i;

	combo = gtk_option_menu_new();
	menu = gtk_menu_new();
	for(i = 0 ; i < sizeof(hashfunctions)/sizeof(hashfunctions[0]) ; i++){
		GtkWidget *item = gtk_menu_item_new_with_label(hashfunctions[i]);
		struct MenuHelper *mh = g_malloc(sizeof(*mh));
		gtk_menu_append(menu,item);
		if(!g_ascii_strncasecmp(app->digesttype,hashfunctions[i],MIN(strlen(app->digesttype),strlen(hashfunctions[i])))){
			gtk_menu_set_active(GTK_MENU(menu),i);
		}
		mh->app = app;
		mh->hashindex = i;
		g_signal_connect(G_OBJECT(item),"activate",G_CALLBACK(digest_changed),mh);
	}
	gtk_option_menu_set_menu(GTK_OPTION_MENU(combo),menu);


	return combo;
}


#endif
int main(int argc, char *argv[])
{
	struct GHApp app;
	GtkWidget *tmp,*tmp1,*mainhbox;

	gtk_init(&argc,&argv);
	parse_options(argc,argv,&app);
	OpenSSL_add_all_digests();

	app.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(app.window),"gHasher");
	g_object_set(G_OBJECT(app.window),"border-width",7,NULL);
	g_signal_connect(G_OBJECT(app.window),"delete-event",G_CALLBACK(gtk_main_quit),NULL);

	mainhbox = gtk_vbox_new(FALSE,3);

	tmp = gtk_label_new("Filename");
	gtk_misc_set_alignment(GTK_MISC(tmp),0.0,0.5);
	gtk_box_pack_start(GTK_BOX(mainhbox),tmp,FALSE,FALSE,0);

	app.filebutton = gtk_button_new_with_mnemonic("...");
	g_signal_connect(G_OBJECT(app.filebutton),"clicked",G_CALLBACK(filenamebutton_clicked),&app);
	tmp1 = gtk_hbox_new(FALSE,0);

	app.fileentry= gtk_entry_new();
	gtk_box_pack_start(GTK_BOX(tmp1),app.fileentry,TRUE,TRUE,0);
	gtk_box_pack_start(GTK_BOX(tmp1),app.filebutton,FALSE,FALSE,0);
	gtk_box_pack_start(GTK_BOX(mainhbox),tmp1,TRUE,FALSE,0);
	
	tmp1 = gtk_hbox_new(FALSE,0);
	tmp = gtk_label_new("Hash ");
	gtk_misc_set_alignment(GTK_MISC(tmp),0.0,0.5); //make text appear at the left, not the middle
	gtk_box_pack_start(GTK_BOX(tmp1),tmp,FALSE,FALSE,0);
	app.digestdropdown = hashdropdown(&app);
	gtk_box_pack_start(GTK_BOX(tmp1),app.digestdropdown,FALSE,FALSE,0);
	gtk_box_pack_start(GTK_BOX(mainhbox),tmp1,FALSE,FALSE,0);

	app.sumentry= gtk_entry_new();
	gtk_entry_set_width_chars(GTK_ENTRY(app.sumentry),40);
	g_object_set(G_OBJECT(app.sumentry),"editable",FALSE,NULL);
	gtk_box_pack_start(GTK_BOX(mainhbox),app.sumentry,FALSE,FALSE,0);

	app.progress = gtk_progress_bar_new();
	gtk_progress_bar_set_text(GTK_PROGRESS_BAR(app.progress),"..."); //set initial som text, else it resizes when we do it later.
	gtk_box_pack_start(GTK_BOX(mainhbox),app.progress,TRUE,FALSE,0);

	tmp = gtk_hbutton_box_new();
	gtk_button_box_set_layout(GTK_BUTTON_BOX(tmp),GTK_BUTTONBOX_END);
	app.actionbutton = gtk_button_new_with_mnemonic(actions[action].name);
	g_signal_connect(G_OBJECT(app.actionbutton),"clicked",G_CALLBACK(actionbutton_clicked),&app);
	gtk_box_pack_end(GTK_BOX(tmp),app.actionbutton,FALSE,FALSE,0);
	gtk_box_pack_start(GTK_BOX(mainhbox),tmp,TRUE,FALSE,0);

	gtk_widget_show_all(mainhbox);
	gtk_container_add(GTK_CONTAINER(app.window),mainhbox);
	gtk_widget_show(app.window);
	setup_window_icon(app.window);
	if(app.cmdfilename != NULL){
		gtk_init_add(start_hash,&app);
	}
	gtk_main();

	return 0;
}
